#include "wino_macro.h"
#include <ap_int.h>
#include <cstdlib>




#ifndef _WINO_IO_HPP_
#define _WINO_IO_HPP_

template<int port_idx>
void load_input_row_from_ddr(
		ap_uint<128>* DDR_port,
		ap_uint<16> input_buffer[INBUFFER_WIDTH][INPUT_BUFFER_DEPTH],
		ap_uint<16> inheight,
		ap_uint<16> inwidth,
		ap_uint<16> inwidth_align8,
		ap_uint<16> indepth_align8,
		ap_uint<16> group_indepth_x_inwidth_align8_by8,
		ap_uint<16> group_indepth_offset_x_inwidth_align8_by8,
		ap_uint<16> inwidth_ceildiv_inbufferwidth,
		ap_uint<16> buffer_address_mid_increment_step,
		ap_uint<16> input_load_burst_length,
		ap_uint<1> row_address_bitnumber_flag,
		ap_int<16> loaded_row_number,
		ap_int<16> row_load_bound)
{
	ap_int<16> row_idx;
	ap_uint<14> loaded_row_number_by4;
	ap_uint<2> loaded_row_number_mod4;


	(loaded_row_number_by4,loaded_row_number_mod4)=loaded_row_number;
	if( loaded_row_number_mod4 <= port_idx )
		row_idx=(loaded_row_number_by4, (ap_uint<2> ) port_idx);
	else
		row_idx=( (ap_uint<14>)(loaded_row_number_by4+1), (ap_uint<2> ) port_idx);


	if(row_idx >= inheight || row_idx >= row_load_bound) return;



	#if DEBUG_FILE_PRINT
		printf("Loading row_idx %d, %d\n", (int) row_idx, port_idx);
	#endif

	#pragma HLS interface m_axi port=DDR_port depth=200704
	#pragma HLS array_partition variable = input_buffer dim=1 complete

	ap_uint<16> register_matrix[8][8];
	#pragma HLS array_partition variable=register_matrix dim=1 complete
	#pragma HLS array_partition variable=register_matrix dim=2 complete

	ap_uint<1> row_pingpong_bit1 = row_idx[INBUFFER_HEIGHT_BITWIDTH];
	ap_uint<2> row_pingpong_bit2 = row_idx.range(INBUFFER_HEIGHT_BITWIDTH+1, INBUFFER_HEIGHT_BITWIDTH);


	ap_uint<INPUT_BUFFER_DEPTH_BITWIDTH - 1 - INDEPTH_MINITILE_SIZE_BITWIDTH> buffer_address_mid=0;
	ap_uint<INPUT_BUFFER_DEPTH_BITWIDTH - 1 - INDEPTH_MINITILE_SIZE_BITWIDTH> buffer_address_mid_offset=0;


	printf("row_address_bitnumber_flag %d\n", (int) row_address_bitnumber_flag);
	#if DEBUG_FILE_PRINT
		if(row_address_bitnumber_flag)
		{
			if( inwidth_ceildiv_inbufferwidth*indepth_align8 < (1<<(INPUT_BUFFER_DEPTH_BITWIDTH - 1)) )
			{
				printf("ERROR: in sufficient IN BUFFER DEPTH, expected depth=[%d x %d], reald depth [%d]\n", (int) inwidth_ceildiv_inbufferwidth, (int) indepth_align8, (1<<(INPUT_BUFFER_DEPTH_BITWIDTH - 1)) );
				exit(-1);
			}
		}
		else
		{
			if( inwidth_ceildiv_inbufferwidth*indepth_align8 < (1<<(INPUT_BUFFER_DEPTH_BITWIDTH - 2)) )
			{
				printf("ERROR: in sufficient IN BUFFER DEPTH, expected depth=[%d x %d], reald depth [%d]\n", (int) inwidth_ceildiv_inbufferwidth, (int) indepth_align8, (1<<(INPUT_BUFFER_DEPTH_BITWIDTH - 2)) );
				exit(-1);
			}
		}
		
	#endif



	ap_uint<INDEPTH_MINITILE_SIZE_BITWIDTH> buffer_address_mini_tile=0;

	ap_uint<3> depth_idx8=0;

	ap_uint<8> read_inwidth_counter=1;
	ap_uint<8> write_inwidth_counter=1;

	ap_uint<32> ddr_address_offset = row_idx * group_indepth_x_inwidth_align8_by8 + group_indepth_offset_x_inwidth_align8_by8;

	ap_uint<128>* DDR_offset = DDR_port + ddr_address_offset;
	
	ap_uint<1> left_down_flag=1;

	ap_uint<INBUFFER_WIDTH_BITWIDTH-3> bank_split_idx=0;

	for(ap_uint<16> cycle=0;cycle<input_load_burst_length;cycle++)
	{
		#pragma HLS pipeline
		ap_uint<128> DDR_data;

		DDR_data = DDR_offset[cycle];
	
		ap_uint<16> DDR_data_divided[8];
		#pragma HLS array_partition variable = DDR_data_divided complete

		for(int i=0;i<8;i++)
		{
		#pragma HLS unroll
			DDR_data_divided[i]=DDR_data.range(i*16+15,i*16);
		}

		if(read_inwidth_counter == inwidth_align8)
		{
			read_inwidth_counter=1;
		}
		else
		{
			read_inwidth_counter++;
		}
		

		


		ap_uint<16> bank_data[8];
		#pragma HLS array_partition variable = bank_data complete

		if(left_down_flag)
		{
			for(int i=0;i<8;i++)
			{
			#pragma HLS unroll
				bank_data[i]=register_matrix[i][0];
			}
			for(int i=0;i<7;i++)
			{
			#pragma HLS unroll
				for(int j=0;j<8;j++)
				{
				#pragma HLS unroll
					register_matrix[j][i]=register_matrix[j][i+1];
				}
			}
			for(int i=0;i<8;i++)
			{ 
			#pragma HLS unroll
				register_matrix[i][7]=DDR_data_divided[i];
			}
		}
		else
		{
			for(int i=0;i<8;i++)
			{
			#pragma HLS unroll
				bank_data[i]=register_matrix[0][i];
			}

			for(int i=0;i<7;i++)
			{
			#pragma HLS unroll
				for(int j=0;j<8;j++)
				{
				#pragma HLS unroll
					register_matrix[i][j]=register_matrix[i+1][j];
				}
			}
			for(int i=0;i<8;i++)
			{ 
			#pragma HLS unroll
				register_matrix[7][i]=DDR_data_divided[i];
			}
		}
		if(depth_idx8==7)
			left_down_flag=~left_down_flag;





		if(cycle>=8)
		{
			ap_uint<INPUT_BUFFER_DEPTH_BITWIDTH> buffer_addr; 
			ap_uint<INPUT_BUFFER_DEPTH_BITWIDTH - 1 - INDEPTH_MINITILE_SIZE_BITWIDTH> buffer_address_mid_rowbit1_offset;
			ap_uint<INPUT_BUFFER_DEPTH_BITWIDTH - 2 - INDEPTH_MINITILE_SIZE_BITWIDTH> buffer_address_mid_rowbit2_offset;
			
			buffer_address_mid_rowbit1_offset=buffer_address_mid_offset;
			buffer_address_mid_rowbit2_offset=buffer_address_mid_offset;

			if(row_address_bitnumber_flag)
			{
				buffer_addr=(row_pingpong_bit1,buffer_address_mid_rowbit1_offset,buffer_address_mini_tile);
			}
			else
			{
				buffer_addr=(row_pingpong_bit2,buffer_address_mid_rowbit2_offset,buffer_address_mini_tile);
			}
			

			input_buffer[bank_split_idx*8+0][buffer_addr]=bank_data[0];
			input_buffer[bank_split_idx*8+1][buffer_addr]=bank_data[1];
			input_buffer[bank_split_idx*8+2][buffer_addr]=bank_data[2];
			input_buffer[bank_split_idx*8+3][buffer_addr]=bank_data[3];
			input_buffer[bank_split_idx*8+4][buffer_addr]=bank_data[4];
			input_buffer[bank_split_idx*8+5][buffer_addr]=bank_data[5];
			input_buffer[bank_split_idx*8+6][buffer_addr]=bank_data[6];
			input_buffer[bank_split_idx*8+7][buffer_addr]=bank_data[7];

			if(write_inwidth_counter == inwidth_align8)
			{
				buffer_address_mid += buffer_address_mid_increment_step;
			}
			else if( write_inwidth_counter.range(INBUFFER_WIDTH_BITWIDTH-1,0) == 0 )
			{
				buffer_address_mid++;
			}

			if( write_inwidth_counter.range(2,0) == 0)
			{
				buffer_address_mid_offset = buffer_address_mid;
			}
			else if(buffer_address_mini_tile == INDEPTH_MINITILE_SIZE-1)
			{
				buffer_address_mid_offset += inwidth_ceildiv_inbufferwidth;
			}

			buffer_address_mini_tile++;

			if(write_inwidth_counter == inwidth_align8)
			{
				bank_split_idx=0;
			}
			else if(write_inwidth_counter.range(2,0)==0)
			{
				bank_split_idx++;
			}
				
			if(write_inwidth_counter == inwidth_align8)
			{
				write_inwidth_counter=1;
			}
			else
			{
				write_inwidth_counter++;
			}
		}
		depth_idx8++;

	}

	for(ap_uint<16> cycle=0;cycle<8;cycle++)
	{
		#pragma HLS pipeline

		ap_uint<16> bank_data[8];
		#pragma HLS array_partition variable = bank_data complete

		if(left_down_flag)
		{
			for(int i=0;i<8;i++)
			{
			#pragma HLS unroll
				bank_data[i]=register_matrix[i][0];
			}
			for(int i=0;i<7;i++)
			{
			#pragma HLS unroll
				for(int j=0;j<8;j++)
				{
				#pragma HLS unroll
					register_matrix[j][i]=register_matrix[j][i+1];
				}
			}
		}
		else
		{
			for(int i=0;i<8;i++)
			{
			#pragma HLS unroll
				bank_data[i]=register_matrix[0][i];
			}

			for(int i=0;i<7;i++)
			{
			#pragma HLS unroll
				for(int j=0;j<8;j++)
				{
				#pragma HLS unroll
					register_matrix[i][j]=register_matrix[i+1][j];
				}
			}
		}
			
		ap_uint<INPUT_BUFFER_DEPTH_BITWIDTH> buffer_addr; 
		ap_uint<INPUT_BUFFER_DEPTH_BITWIDTH - 1 - INDEPTH_MINITILE_SIZE_BITWIDTH> buffer_address_mid_rowbit1_offset;
		ap_uint<INPUT_BUFFER_DEPTH_BITWIDTH - 2 - INDEPTH_MINITILE_SIZE_BITWIDTH> buffer_address_mid_rowbit2_offset;
		
		buffer_address_mid_rowbit1_offset=buffer_address_mid_offset;
		buffer_address_mid_rowbit2_offset=buffer_address_mid_offset;

		if(row_address_bitnumber_flag)
		{
			buffer_addr=(row_pingpong_bit1,buffer_address_mid_rowbit1_offset,buffer_address_mini_tile);
		}
		else
		{
			buffer_addr=(row_pingpong_bit2,buffer_address_mid_rowbit2_offset,buffer_address_mini_tile);
		}
		input_buffer[bank_split_idx*8+0][buffer_addr]=bank_data[0];
		input_buffer[bank_split_idx*8+1][buffer_addr]=bank_data[1];
		input_buffer[bank_split_idx*8+2][buffer_addr]=bank_data[2];
		input_buffer[bank_split_idx*8+3][buffer_addr]=bank_data[3];
		input_buffer[bank_split_idx*8+4][buffer_addr]=bank_data[4];
		input_buffer[bank_split_idx*8+5][buffer_addr]=bank_data[5];
		input_buffer[bank_split_idx*8+6][buffer_addr]=bank_data[6];
		input_buffer[bank_split_idx*8+7][buffer_addr]=bank_data[7];

 		if(buffer_address_mini_tile == INDEPTH_MINITILE_SIZE-1)
		{
			buffer_address_mid_offset += inwidth_ceildiv_inbufferwidth;
		}
		buffer_address_mini_tile++;
	}
}




//template<int DUMMY>
void load_input_rowtile_from_ddr(
		ap_uint<128>* DDR_port1,
        ap_uint<128>* DDR_port2,
        ap_uint<128>* DDR_port3,
        ap_uint<128>* DDR_port4,
		ap_uint<16> input_buffer[INBUFFER_HEIGHT][INBUFFER_WIDTH][INPUT_BUFFER_DEPTH],
		ap_uint<16> inheight,
		ap_uint<16> inwidth,
		ap_uint<16> stride,
        ap_uint<16> pad_size,
		ap_uint<16> inwidth_align8,
		ap_uint<16> indepth_align8,
		ap_uint<16> group_indepth_x_inwidth_align8_by8,
		ap_uint<16> group_indepth_offset_x_inwidth_align8_by8,
		ap_uint<16> inwidth_ceildiv_inbufferwidth,
		ap_uint<16> buffer_address_mid_increment_step,
		ap_uint<16> input_load_burst_length,
		ap_uint<1>  row_address_bitnumber_flag,
		ap_uint<16> out_rowstep,
		ap_uint<16> next_start_output_row,
        ap_uint<1> first_flag
)
{
	#pragma HLS array_partition variable =input_buffer dim=1 complete
	#pragma HLS array_partition variable =input_buffer dim=2 complete
    static int loaded_input_row_number;
    if(first_flag)
        loaded_input_row_number = 0;
    
    int required_loaded_input_row_number = (next_start_output_row+out_rowstep)*stride + pad_size;

    while( required_loaded_input_row_number > loaded_input_row_number)
    {
        if( loaded_input_row_number%8/4 == 0)
        {

            load_input_row_from_ddr<0>(
                DDR_port1,
                input_buffer[0],
				inheight,
				inwidth,
				inwidth_align8,
				indepth_align8,
				group_indepth_x_inwidth_align8_by8,
				group_indepth_offset_x_inwidth_align8_by8,
				inwidth_ceildiv_inbufferwidth,
				buffer_address_mid_increment_step,
				input_load_burst_length,
				row_address_bitnumber_flag,
                loaded_input_row_number,
				required_loaded_input_row_number
            );


            load_input_row_from_ddr<1>(
                DDR_port2,
                input_buffer[1],
				inheight,
				inwidth,
				inwidth_align8,
				indepth_align8,
				group_indepth_x_inwidth_align8_by8,
				group_indepth_offset_x_inwidth_align8_by8,
				inwidth_ceildiv_inbufferwidth,
				buffer_address_mid_increment_step,
				input_load_burst_length,
				row_address_bitnumber_flag,
                loaded_input_row_number,
				required_loaded_input_row_number
            );


            load_input_row_from_ddr<2>(
                DDR_port3,
                input_buffer[2],
				inheight,
				inwidth,
				inwidth_align8,
				indepth_align8,
				group_indepth_x_inwidth_align8_by8,
				group_indepth_offset_x_inwidth_align8_by8,
				inwidth_ceildiv_inbufferwidth,
				buffer_address_mid_increment_step,
				input_load_burst_length,
				row_address_bitnumber_flag,
                loaded_input_row_number,
				required_loaded_input_row_number
            );

            load_input_row_from_ddr<3>(
                DDR_port4,
                input_buffer[3],
				inheight,
				inwidth,
				inwidth_align8,
				indepth_align8,
				group_indepth_x_inwidth_align8_by8,
				group_indepth_offset_x_inwidth_align8_by8,
				inwidth_ceildiv_inbufferwidth,
				buffer_address_mid_increment_step,
				input_load_burst_length,
				row_address_bitnumber_flag,
                loaded_input_row_number,
				required_loaded_input_row_number
            );

        }
        else
        {

            load_input_row_from_ddr<0>(
                DDR_port1,
                input_buffer[4],
				inheight,
				inwidth,
				inwidth_align8,
				indepth_align8,
				group_indepth_x_inwidth_align8_by8,
				group_indepth_offset_x_inwidth_align8_by8,
				inwidth_ceildiv_inbufferwidth,
				buffer_address_mid_increment_step,
				input_load_burst_length,
				row_address_bitnumber_flag,
                loaded_input_row_number,
				required_loaded_input_row_number
            );

            load_input_row_from_ddr<1>(
                DDR_port2,
                input_buffer[5],
				inheight,
				inwidth,
				inwidth_align8,
				indepth_align8,
				group_indepth_x_inwidth_align8_by8,
				group_indepth_offset_x_inwidth_align8_by8,
				inwidth_ceildiv_inbufferwidth,
				buffer_address_mid_increment_step,
				input_load_burst_length,
				row_address_bitnumber_flag,
                loaded_input_row_number,
				required_loaded_input_row_number
            );


            load_input_row_from_ddr<2>(
                DDR_port3,
                input_buffer[6],
				inheight,
				inwidth,
				inwidth_align8,
				indepth_align8,
				group_indepth_x_inwidth_align8_by8,
				group_indepth_offset_x_inwidth_align8_by8,
				inwidth_ceildiv_inbufferwidth,
				buffer_address_mid_increment_step,
				input_load_burst_length,
				row_address_bitnumber_flag,
                loaded_input_row_number,
				required_loaded_input_row_number
            );

            load_input_row_from_ddr<3>(
                DDR_port4,
                input_buffer[7],
				inheight,
				inwidth,
				inwidth_align8,
				indepth_align8,
				group_indepth_x_inwidth_align8_by8,
				group_indepth_offset_x_inwidth_align8_by8,
				inwidth_ceildiv_inbufferwidth,
				buffer_address_mid_increment_step,
				input_load_burst_length,
				row_address_bitnumber_flag,
                loaded_input_row_number,
				required_loaded_input_row_number
            );
        }
        loaded_input_row_number+=4;
    }
	loaded_input_row_number=required_loaded_input_row_number;
	#if DEBUG_FILE_PRINT
	fflush(stdout);
	#endif
}



//template<int DUMMY>
void write_output_to_DDR(
	ap_uint<128>* DDR_port0,
	ap_uint<128>* DDR_port1,
	ap_uint<128>* DDR_port2,
	ap_uint<128>* DDR_port3,
	ap_uint<36> output_buffer[16][16][1024],
	int outdepth_ceildiv8,
	int start_out_row_idx,
	int end_row,
	int out_height_aligned4,	
	int out_width,
	int wino_output_tile_size,
	int row_tile_number,
	int wino_flag,
	int final_right_shift
	#if DEBUG_FILE_PRINT

	#endif
	
	)
{
#pragma HLS interface m_axi port = DDR_port0
#pragma HLS interface m_axi port = DDR_port1
#pragma HLS interface m_axi port = DDR_port2
#pragma HLS interface m_axi port = DDR_port3

#pragma HLS array_partition variable = output_buffer dim=1 complete
#pragma HLS array_partition variable = output_buffer dim=2 complete

	ap_uint<1> valid_output_flag[4];
	
	int row_offset0 = (start_out_row_idx+0)*out_width; 
	int row_offset1 = (start_out_row_idx+1)*out_width; 
	int row_offset2 = (start_out_row_idx+2)*out_width; 
	int row_offset3 = (start_out_row_idx+3)*out_width;

	int plane_offset_step = out_width*out_height_aligned4;
	int depth_base_offset = 0;
	

	for(int depth_idx=0;depth_idx<outdepth_ceildiv8;depth_idx++)
	{
		
		ap_uint<128>* DDR_offset0= DDR_port0+depth_idx*out_width*out_height_aligned4+row_offset0;
		ap_uint<128>* DDR_offset1= DDR_port1+depth_idx*out_width*out_height_aligned4+row_offset1;
		ap_uint<128>* DDR_offset2= DDR_port2+depth_idx*out_width*out_height_aligned4+row_offset2;
		ap_uint<128>* DDR_offset3= DDR_port3+depth_idx*out_width*out_height_aligned4+row_offset3;
		int buffer_address = depth_idx*row_tile_number;
		ap_uint<4> buffer_address_incre_cycle_count=0;
		for(ap_int<16> out_col_idx=0;out_col_idx<out_width;out_col_idx++)
		{
#pragma HLS pipeline


			ap_uint<36> output_buffer_port_wire[16][16];
#pragma HLS array_partition variable = output_buffer_port_wire  complete
			for(int i=0;i<16;i++){
			#pragma HLS unroll
			for(int j=0;j<16;j++){
			#pragma HLS unroll
				output_buffer_port_wire[i][j]=output_buffer[i][j][buffer_address];
			}
			}

			ap_uint<36> output_buffer_value_wire[8][4][2][4];
#pragma HLS array_partition variable = output_buffer_value_wire dim=1   complete
#pragma HLS array_partition variable = output_buffer_value_wire dim=2   complete
#pragma HLS array_partition variable = output_buffer_value_wire dim=3   complete
#pragma HLS array_partition variable = output_buffer_value_wire dim=4   complete
			for(int i=0;i<8;i++){
			#pragma HLS unroll	
			for(int j=0;j<2;j++){
			#pragma HLS unroll	
			for(int k=0;k<4;k++){
			#pragma HLS unroll	
			for(int l=0;l<4;l++){
			#pragma HLS unroll	
				output_buffer_value_wire[i][k][j][l]=output_buffer_port_wire[j*8+i][k*4+l];
			}
			}
			}
			}

			

			ap_uint<36> write_back_wire[8][4];
#pragma HLS array_partition variable = write_back_wire  complete
			ap_uint<2> pix_sel = out_col_idx.range(1,0);
			ap_uint<1> tile_sel = out_col_idx[2];

			for(int i=0;i<8;i++){
			#pragma HLS unroll	
			for(int j=0;j<4;j++){
			#pragma HLS unroll	
				write_back_wire[i][j]=output_buffer_value_wire[i][j][tile_sel][pix_sel];
			}
			}

			ap_int<8> write_back_wire_batch0[8][4];
			ap_int<8> write_back_wire_batch1[8][4];
#pragma HLS array_partition variable = write_back_wire_batch0  complete
#pragma HLS array_partition variable = write_back_wire_batch1  complete
			for(int i=0;i<8;i++){
				#pragma HLS unroll
			for(int j=0;j<4;j++){
				#pragma HLS unroll
				write_back_wire_batch0[i][j]= (write_back_wire[i][j].range(17,0)) >> final_right_shift;
				write_back_wire_batch1[i][j]= (write_back_wire[i][j].range(35,18)) >> final_right_shift;
			}
			}

			ap_uint<128> write_back_128bit3;
			ap_uint<128> write_back_128bit2;
			ap_uint<128> write_back_128bit1;
			ap_uint<128> write_back_128bit0;

			for(int i=0;i<8;i++)
			{
				#pragma HLS unroll
				write_back_128bit0.range(i*16+15,i*16) = ( write_back_wire_batch1[i][0],write_back_wire_batch0[i][0]);
				write_back_128bit1.range(i*16+15,i*16) = ( write_back_wire_batch1[i][1],write_back_wire_batch0[i][1]);
				write_back_128bit2.range(i*16+15,i*16) = ( write_back_wire_batch1[i][2],write_back_wire_batch0[i][2]);
				write_back_128bit3.range(i*16+15,i*16) = ( write_back_wire_batch1[i][3],write_back_wire_batch0[i][3]);
			}

			
			DDR_offset3[out_col_idx]=write_back_128bit3;
			DDR_offset2[out_col_idx]=write_back_128bit2;
			DDR_offset1[out_col_idx]=write_back_128bit1;
			DDR_offset0[out_col_idx]=write_back_128bit0;
			if(buffer_address_incre_cycle_count == wino_output_tile_size*2-1)
			{
				buffer_address_incre_cycle_count=0;
				buffer_address ++;
			}
			else
			{
				buffer_address_incre_cycle_count++;
			}
			

		}
	}

}



#endif
